/*
 * ======================================================================
 * Dimension.java:
 *
 * Written by: Mark Austin
 * ======================================================================
 */

package whistle;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.lang.Math;

public class Dimension {
   double  scaleFactor;  // scale/conversion factor 
   double  lengthExpnt;  // length                 
   double    massExpnt;  // mass                  
   double    timeExpnt;  // time  
   double    tempExpnt;  // temperature
   double  radianExpnt;  // planar angle 
   DimensionType type;   // US or SI units 

   // ================================================
   // Constructor methods ...
   // ================================================

   public Dimension() {
      this.scaleFactor = 1.0;
      this.type        = DimensionType.SI;
   }

   public Dimension( double scaleFactor, double  lengthExpnt,
                     double massExpnt,   double timeExpnt, double  tempExpnt,
                     double radianExpnt, DimensionType type ) {
      this.scaleFactor = scaleFactor;  // scale factor ..
      this.lengthExpnt = lengthExpnt;  // length                 
      this.massExpnt   = massExpnt;    // mass                 
      this.timeExpnt   = timeExpnt;    // time 
      this.tempExpnt   = tempExpnt;    // temperature
      this.radianExpnt = radianExpnt;  // radian
      this.type        = type;         // dimensions type ...
   }

   // ================================================
   // Set/get dimensions type ...
   // ================================================

   public void setUnitsType ( DimensionType type ) {
      this.type = type;
   }

   public DimensionType getUnitsType () {
      return type;
   }

   // ================================================
   // Set/get scale factor and dimension exponents ...
   // ================================================

   public void setScaleFactor( double scaleFactor ) {
      this.scaleFactor = scaleFactor;
   }

   public double getScaleFactor() {
      return scaleFactor;
   }

   // Length exponents ....

   public void setLengthExponent( double lengthExpnt ) {
      this.lengthExpnt = lengthExpnt;
   }

   public double getLengthExponent() {
      return lengthExpnt;
   }

   // Mass exponents ....

   public void setMassExponent( double massExpnt ) {
      this.massExpnt = massExpnt;
   }

   public double getMassExponent() {
      return massExpnt;
   }

   // Time exponents ....

   public void setTimeExponent( double timeExpnt ) {
      this.timeExpnt = timeExpnt;
   }

   public double getTimeExponent() {
      return timeExpnt;
   }

   // Temperature exponents ....

   public void setTempExponent( double tempExpnt ) {
      this.tempExpnt = tempExpnt;
   }

   public double getTempExponent() {
      return tempExpnt; 
   }

   // Radian exponents ....

   public void setRadianExponent( double radianExpnt ) {
      this.radianExpnt = radianExpnt;
   }

   public double getRadianExponent() {
      return radianExpnt;
   }

   // ======================================================
   // Computational units and dimensions ...
   // ======================================================

   public boolean sameDimension( Dimension temp ) {
      boolean result = false;

      if(( Math.abs( this.scaleFactor - temp.scaleFactor ) < 0.0000000000001 ) &&
          (this.lengthExpnt  == temp.lengthExpnt) &&
          (this.massExpnt    == temp.massExpnt)   &&
          (this.timeExpnt    == temp.timeExpnt)   &&
          (this.tempExpnt    == temp.tempExpnt)   &&
          (this.radianExpnt  == temp.radianExpnt) &&
          (this.type         == temp.type) )  {
           result = true;
      }

      return result;
   }

   public boolean sameUnits( Dimension temp ) {
      boolean result = false;

      if((this.lengthExpnt  == temp.lengthExpnt) &&
         (this.massExpnt    == temp.massExpnt)   &&
         (this.timeExpnt    == temp.timeExpnt)   &&
         (this.tempExpnt    == temp.tempExpnt)   &&
         (this.radianExpnt  == temp.radianExpnt) &&
         (this.type         == temp.type) )  {
          result = true;
      }
    
      return result;
   }

   public boolean compatibleUnits( Dimension temp ) {
      boolean result = false;

      if((this.lengthExpnt  == temp.lengthExpnt) &&
         (this.massExpnt    == temp.massExpnt)   &&
         (this.timeExpnt    == temp.timeExpnt)   &&
         (this.tempExpnt    == temp.tempExpnt)   &&
         (this.radianExpnt  == temp.radianExpnt) ) {
          result = true;
      }
    
      return result;
   }

   public boolean isDimensionless() {
      boolean result = false;

      if((this.lengthExpnt  == 0.0) &&
         (this.massExpnt    == 0.0)   &&
         (this.timeExpnt    == 0.0)   &&
         (this.tempExpnt    == 0.0)   &&
         (this.radianExpnt  == 0.0) ) {
          result = true;
      }

      return result;
   }

   // ======================================================
   // Computational methods ...
   // ======================================================

   public Dimension scale( int n ) {
      Dimension result = new Dimension();

      result.scaleFactor = Math.pow (this.scaleFactor, (double) n );
      result.lengthExpnt = n * this.lengthExpnt;
      result.massExpnt   = n * this.massExpnt; 
      result.timeExpnt   = n * this.timeExpnt;
      result.tempExpnt   = n * this.tempExpnt;
      result.radianExpnt = n * this.radianExpnt;
      result.type        = this.type;  

      return result;
   }

   public Dimension scale( double n ) {
      Dimension result = new Dimension();

      result.scaleFactor = Math.pow (this.scaleFactor, (double) n );
      result.lengthExpnt = n * this.lengthExpnt;
      result.massExpnt   = n * this.massExpnt; 
      result.timeExpnt   = n * this.timeExpnt;
      result.tempExpnt   = n * this.tempExpnt;
      result.radianExpnt = n * this.radianExpnt;
      result.type        = this.type;  

      return result;
   }

   public Dimension add( Dimension a ) {
      Dimension result = new Dimension();

      result.scaleFactor = this.scaleFactor * a.scaleFactor;
      result.lengthExpnt = this.lengthExpnt + a.lengthExpnt;
      result.massExpnt   = this.massExpnt   + a.massExpnt;
      result.timeExpnt   = this.timeExpnt   + a.timeExpnt;
      result.tempExpnt   = this.tempExpnt   + a.tempExpnt;
      result.radianExpnt = this.radianExpnt + a.radianExpnt;
      result.type        = this.type;  

      return result;
   }

   public Dimension sub( Dimension a ) {
      Dimension result = new Dimension();

      result.scaleFactor = this.scaleFactor / a.scaleFactor;
      result.lengthExpnt = this.lengthExpnt - a.lengthExpnt;
      result.massExpnt   = this.massExpnt   - a.massExpnt;
      result.timeExpnt   = this.timeExpnt   - a.timeExpnt;
      result.tempExpnt   = this.tempExpnt   - a.tempExpnt;
      result.radianExpnt = this.radianExpnt - a.radianExpnt;
      result.type        = this.type;  

      return result;
   }

   public Dimension pow( int n ) {
      Dimension result = new Dimension();

      result.scaleFactor = Math.pow( this.scaleFactor, (double) n );
      result.lengthExpnt = n * this.lengthExpnt;
      result.massExpnt   = n * this.massExpnt; 
      result.timeExpnt   = n * this.timeExpnt; 
      result.tempExpnt   = n * this.tempExpnt; 
      result.radianExpnt = n * this.radianExpnt;
      result.type        = this.type;  

      return result;
   }

   // =====================================================
   // Copy dimensions from "a" to "this" ....
   // =====================================================

   public void copy( Dimension d02 ) {
      this.scaleFactor = d02.scaleFactor;
      this.lengthExpnt = d02.lengthExpnt;
      this.massExpnt   = d02.massExpnt; 
      this.timeExpnt   = d02.timeExpnt; 
      this.tempExpnt   = d02.tempExpnt; 
      this.radianExpnt = d02.radianExpnt;
      this.type        = d02.type;  
   }

   // =====================================================
   // Zero out units and name ....
   // =====================================================

   public void zero() {
      this.scaleFactor = 1.0;
      this.lengthExpnt = 0.0;
      this.massExpnt   = 0.0;
      this.timeExpnt   = 0.0;
      this.tempExpnt   = 0.0;
      this.radianExpnt = 0.0;
      if(this.type != DimensionType.SI && this.type != DimensionType.US )
         this.type = DimensionType.SI_US;
   }

   // Set default units according to a name ...

   public void defaultUnits( String name ) {
   }

   // Convert units type ....

   public Dimension unitsTypeConvert( DimensionType type ) {
      Dimension result = null;

      if( this.type != type || type != DimensionType.SI_US ) {
         result = new Dimension();
         result.setUnitsType( type );
         // result.setName( null ); 
         // UnitsSimplify(d);
      } 

      return result;
   }

   // Simplify units ...

   public Dimension simplify() {
      Dimension temp = new Dimension();
      return temp;
   }

   public Dimension radUnitsSimplify() {
      Dimension temp = new Dimension();
      return temp;
   }

   // ------------------------------------------------------------
   // Create clone (i.e., deep copy) of a Dimension object ...
   // ------------------------------------------------------------

   public Object clone() {
      try {
         Dimension dim = (Dimension) super.clone();
         return dim;
      } catch (CloneNotSupportedException e) {
         return null;
      }
   }

   // ============================================================
   // String representation of dimensions ...
   // ============================================================

   public String toString() {
      StringBuffer buf = new StringBuffer();
      buf.append( String.format("Dimension: scale  = %12.8g\n", scaleFactor ) );
      buf.append( String.format("Exponents: length = %12.8g\n", lengthExpnt ) );
      buf.append( String.format("         :   mass = %12.8g\n", massExpnt ) );
      buf.append( String.format("         :   time = %12.8g\n", timeExpnt ) );
      buf.append( String.format("         :   temp = %12.8g\n", tempExpnt ) );
      buf.append( String.format("         : radian = %12.8g\n", radianExpnt ) );
      switch ( type ) {
          case SI:
               buf.append( String.format("         :   type = SI\n" ));
               break;
          case US:
               buf.append( String.format("         :   type = US\n" ));
               break;
          case SI_US:
               buf.append( String.format("         :   type = SI_US\n" ));
          default:
               break;
      }

      return buf.toString();
   }

   // ============================================================
   // Override equals() and hashCode() methods ....
   // ============================================================

   public boolean equals(Object other) {   

      // See if we're referring to the same objects  

      if (this == other) return true;   

      // See if a null was passed in or an object that isn't a dimension ...

      if (other == null || !(other instanceof Dimension))
          return false;   

      // Systematically compare arguments of each dimension ...

      Dimension othertester = (Dimension) other;  
      boolean compatible = this.sameDimension( othertester );
      return ( compatible );
   }

   public int hashCode() {
      double hash = lengthExpnt * 31;  
      hash = hash + 17*massExpnt + 3*timeExpnt;  
      hash = hash + 2*tempExpnt + radianExpnt;  
      hash = hash + Math.log(scaleFactor);  
      return ((int) Math.abs(hash));
   }  

   // =========================================================
   // Exercise methods in Dimension ....
   // =========================================================

   public static void main( String args [] ) {

      // Part 01: Create and print a simple dimension ....

      System.out.println("");
      System.out.println("*** Exercise methods in whistle/Dimension ...");
      System.out.println("*** ========================================== ");

      Dimension d1 = new Dimension ();
      System.out.println( d1.toString() );

      // Part 02: Build library of well-known unit names and their properties ...

      LibraryDimension lib = new LibraryDimension();
      HashMap library = lib.getStringToLibraryDimen();

      System.out.println("");
      System.out.println("*** Test equality of dimensions ...");
      System.out.println("*** ----------------------------------");
      System.out.println("");

      Dimension d01 = (Dimension) library.get("m");
      Dimension d02 = (Dimension) library.get("mm");

      if ( d01.sameDimension (d02) == true )
         System.out.println("*** Dimension d01 and d02 are equal!");
      else
         System.out.println("*** Dimension d01 and d02 are different!");

      if ( d01.sameUnits (d02) == true )
         System.out.println("*** d01 and d02 have the same units!");
      else
         System.out.println("*** d02 and d02 have different units!");

   }
}
