/*
 *  =======================================================================
 *  Compilation:  javac MatrixSimple.java
 *  Execution:    java MatrixSimple
 *
 *  Basic implementation of an M-by-N matrix and matrix operations
 *
 *  Written by: Mark Austin                                       July 2013
 *  =======================================================================
 */

package whistle;

import java.lang.Math;
import java.util.*;
import java.io.*;
import java.text.*;

public class MatrixSimple implements Matrix {
    private String          sName;  // name of array ...
    private int           iNoRows;  // # rows
    private int        iNoColumns;  // # columns
    private Dimension [] rowUnits;  // Row units buffer ..
    private Dimension [] colUnits;  // Column units buffer ...
    private double[][] data;        // The 2d array
    boolean DEBUG = false;

    private LibraryDimension library = new LibraryDimension();
    private static HashMap d2s = LibraryDimension.getDimenToStringLibrary();
    private static HashMap s2d = LibraryDimension.getStringToLibraryDimen();

    // Constructors for M-by-N matrices of 0's

    public MatrixSimple(int iNoRows, int iNoColumns ) {
        this.iNoRows    = iNoRows;
        this.iNoColumns = iNoColumns;
        this.colUnits   = new Dimension[ iNoColumns ];         // Array of references to column units.
        this.rowUnits   = new Dimension[ iNoRows ];            // Array of references to row units.
        for (int i = 0; i < iNoRows; i = i + 1)
            this.rowUnits[i] = new Dimension();                // Allocate row units objects.
        for (int j = 0; j < iNoColumns; j = j + 1)
            this.colUnits[j] = new Dimension();                // Allocate column units objects.

        this.data       = new double[ iNoRows ][ iNoColumns ]; // Two-dimensional data array.
    }

    public MatrixSimple( String sName, int iNoRows, int iNoColumns ) {
        this.sName      = sName;
        this.iNoRows    = iNoRows;
        this.iNoColumns = iNoColumns;
        this.colUnits   = new Dimension[ iNoColumns ];         // Array of references to column units.
        this.rowUnits   = new Dimension[ iNoRows ];            // Array of references to row units.
        for (int i = 0; i < iNoRows; i = i + 1)
            this.rowUnits[i] = new Dimension();                // Allocate row units objects.
        for (int j = 0; j < iNoColumns; j = j + 1)
            this.colUnits[j] = new Dimension();                // Allocate column units objects.

        this.data       = new double[ iNoRows ][ iNoColumns ]; // Two-dimensional data array.
    }

    // ======================================
    // Set/get variables for data array
    // ======================================

    public void setName ( String sName ) {
       this.sName = sName;
    }

    public String getName() {
       return this.sName;
    }

    public int getNoRows() {
       return iNoRows;
    }

    public int getNoColumns() {
       return iNoColumns;
    }

    public double getData(int i, int j) {
       return data[i][j];
    }

    public void setData(int i, int j, double value ) {
       data[i][j] = value;
    }

    // ======================================
    // Retrieve and set matrix element values
    // ======================================

    public Quantity getElement( int i, int j ) {
        Quantity returnValue;

        if( i < 0 || j < 0 || i >= iNoRows || j >= iNoColumns ) throw new
            RuntimeException("*** In getElement(): Array element index out of range");

        returnValue = new Quantity( data[i][j] );
        returnValue.setDimension( this.rowUnits[i].add ( this.colUnits[j] ) );

        return returnValue;
    }

    public void setElement( int i, int j, Quantity q01 )  {

        // Check element index values are well defined ....

        if( i < 0 || j < 0 || i > (iNoRows-1) || j > (iNoColumns-1) ) {
            System.out.println("*** In whistle/MatrixSimple.setElement() ... ");
            System.out.printf ("*** ERROR: Array name        = %s ... \n", getName() );
            System.out.printf ("*** ERROR: Array index [i,j] = [%d, %d] out of range ... \n", i, j);
            System.out.printf ("*** ERROR: No matrix rows    = %d ... \n", iNoRows );
            System.out.printf ("*** ERROR: No matrix columns = %d ... \n", iNoColumns );
        }

        // By default, matrix element units are stored in the
        // column units buffer ....

        Dimension units = q01.getDimension();
       
        if ( DEBUG == true ) {
           System.out.printf("*** In MatrixSimple.setElement(): [i,j] = [%d, %d]\n", i, j);
           System.out.println( units.toString() );
        }

        if ( this.colUnits[j].isDimensionless() == true ) {
            this.colUnits[j].copy( units );
            this.data[i][j] = q01.getValue();
        } else if ( this.colUnits[j].sameUnits( units ) == true ) {
            double sf01   = this.colUnits[j].getScaleFactor();
            double sf02   = q01.getDimension().getScaleFactor();
            double result = (sf02/sf01)*q01.getValue();

            if ( DEBUG == true ) {
               System.out.printf("*** sf01   = %f\n", sf01 );
               System.out.printf("*** sf02   = %f\n", sf02 );
               System.out.printf("*** result = %f\n", result );
            }

            this.data[i][j] = result;
        } else {
            this.colUnits[j].zero();
        }    
    }

    // ================================================
    // Create matrices based on 2d arrays
    // ================================================

    public MatrixSimple(String sName, double[][] data) {
        this.sName      = sName;
        this.iNoRows    = data.length;
        this.iNoColumns = data[0].length;
        this.data       = new double[ iNoRows ][ iNoColumns ];
        for (int i = 0; i < iNoRows; i++)
            for (int j = 0; j < iNoColumns; j++)
                    this.data[i][j] = data[i][j];
    }

    public MatrixSimple(double[][] data) {
        this.iNoRows    = data.length;
        this.iNoColumns = data[0].length;
        this.data       = new double[ iNoRows ][ iNoColumns ];
        for (int i = 0; i < iNoRows; i++)
            for (int j = 0; j < iNoColumns; j++)
                    this.data[i][j] = data[i][j];
    }

    // ================================================
    // copy constructor
    // ================================================

    public MatrixSimple( MatrixSimple A ) {
        this(A.data);
    }

    // ========================================================================
    // columnUnits() : Write a (1xN) column units buffer into a matrix
    //
    // Usage: Argument is matrix of units to be applied.
    //
    // Syntax:    matrix.columnUnits( [ units_vector ] );
    //
    // Examples : matrix.columnUnits( [ kg ] );
    //            matrix.columnUnits( [ N/m, N/m, rad ] );
    //
    // Notes:
    //
    // 1. When the parent matrix is dimensionless and the number of
    //    columns in the dimensions matrix matches the number of columns
    //    in the matrix, column units will be matched for each matrix column.
    //
    // 2. Column i of matrix m has units compatible with the dimensions     
    //    matrix. Copy units to the matching coluns and rescale elements.                 
    //    Example:
    //
    //    Matrix m01 = Matrix( [10,2] );
    //    m01.columnUnits( [  m ] ); <-- column units will be meters.
    //    m01.columnUnits( [ cm ] ); <-- now column units will be centermeters.
    //
    // 3. Otherwise, the same unit will be applied to all columns.
    //
    // ========================================================================

    public void columnUnits ( Matrix units ) {
       int ID = 1;

       // [a] : Check units flags and get appropriate units ...

       // .... fill in later ....

       // [b] : Get appropriate units ....

       // [c] : Retrieve and check appropriate units ...

       // ===================================================================== 
       // [f] : Initialize non_units row buffer ....
       // ===================================================================== 

       for( int i = 1; i <= this.iNoRows; i++) { }

    }

    // ============================================================================
    // columnUnits() : Write a column units vector to a specified column number.
    //
    // Usage: Argument is matrix of units to be applied.
    //
    // Syntax:    matrix.columnUnits( [ units_vector ], [ number ] );
    //
    // Examples : matrix.columnUnits( [ sec ], [1] );
    //            matrix.columnUnits( [ Pa  ], [2] );
    //
    // Notes:
    //
    // =========================================================================

    public void columnUnits ( Matrix units, Matrix columns ) {
       int ID = 1;

       // [a] : Check units flags and get appropriate units ...

       // .... fill in later ....

       // [b] : Get appropriate units ....

       // [c] : Retrieve and check appropriate units ...

       // ===================================================================== 
       // [f] : Initialize non_units row buffer ....
       // ===================================================================== 

       for( int i = 1; i <= this.iNoRows; i++) { }

    }

    // ====================================================================
    // rowUnits() : Specify row units in a matrix
    // 
    // Usage:  1st argument is matrix of units to be applied.
    //         2nd argument row number for units.
    // 
    // Examples : matrix.rowUnits( [ kg ] );
    //            matrix.rowUnits( [ kg ], [2] );
    // 
    // ====================================================================

    public void rowUnits ( Matrix first ) {
       int ID = 0;

       // [a] : Check units flags and get appropriate units ....

       // [b] : Get appropriate units ...

       // [c] : Retrieve and check appropriate units ....
  
       // ===================================================================== 
       // [d] : Case 1 : Update "row units" in one row only.
       //
       // Two cases exist -- (1.1) Row i of matrix m has no units       
       //                          (1.2) Row i of matrix m has units.    
       //                                Check compatibility with units_m.
       //                                Copy unmits_m to m.        
       //                                                            
       // Examples : stiff = Matrix([2,2]);                           
       //            stiff.rowUnits ( [  sec], [1] );         
       //            stiff.rowUnits ( [sec^2], [2] );          
       //                                                              
       // ===================================================================== 

       if (ID == 1) {
       }

       // ===================================================================== 
       // [e] : Case 2 : Update "units" in multiple matrix rows.               
       //                                                                       
       // Cases -- (2.1) Matrix m and units_m have the same number of rows.      
       //          (2.1.1) : Matrix m has no units. Assign units from units_m to m
       //          (2.1.2) : Matrix m has units. Copy units_m to units of m.     
       //          (2.2) Matrices m and units_m have different number of rows.   
       //          (2.2.1) : matrix m rows have no units. Assign all rows of m   
       //                    with units of units_m                        
       //          (2.2.2) : Assign unit_m to rows of m, where they match.
       //                                                           
       // Examples : stiff = Matrix([3,3]);                          
       //            stiff.rowUnits ( [cm/sec^2], [1] );     
       //            stiff.rowUnits ( [sec] );               
       //            stiff.rowUnits ( [in/sec^2] );           
       //  
       // ===================================================================== 

       if (ID == 2) {

          // CASE[a]: Matrix m's row number is the same as Units Matrix 
          //          units_m's column number                          

          // CASE [b] : No of rows in matrix m and No of 
          //            columns in units_m are different 

       }
    }

    // ====================================================================
    // create and return a random M-by-N matrix with values between 0 and 1
    // ====================================================================

    public static Matrix random( int iNoRows, int iNoColumns ) {
        Matrix A = new MatrixSimple( iNoRows, iNoColumns );
        for (int i = 0; i < iNoRows; i++)
            for (int j = 0; j < iNoColumns; j++)
                A.setData( i, j, Math.random());

        return A;
    }

    // ====================================================================
    // Create and return the N-by-N identity matrix
    // ====================================================================

    public static Matrix identity( int iNoRows ) {

        Matrix I = new MatrixSimple( iNoRows, iNoRows);
        for (int i = 0; i < iNoRows; i++) {
             I.setData( i, i, 1.0 );
        }

        return I;
    }

    // ====================
    // Swap rows i and j
    // ====================

    public void swap(int i, int j) {
        double[] temp = data[i];
        data[i]       = data[j];
        data[j]       = temp;
    }

    // ====================================================================
    // Create diagonal matrix ...
    // ====================================================================

    public Matrix diagonal( Matrix diagonals ) {
        Matrix result = null;
        int iNoRows = diagonals.getNoRows();
        int iNoCols = diagonals.getNoColumns();

        if( iNoRows < 1 || iNoCols < 2 ) {
           System.out.println("*** In SimpleMatrix.diagonal() ... " );
           System.out.println("*** ERROR: Argument matrix too small" );
        }

        if ( iNoRows == 1 && iNoCols > 1 )
           result = new MatrixSimple( iNoCols, iNoCols );
        else
           result = new MatrixSimple( iNoRows, iNoRows );

        // Initialize matrix diagonal elements ....

        // .... tbd ......

        return result;
    }

    // ====================================================================
    // Scale a matrix by a constant ...
    // ====================================================================

    public Matrix scale( Quantity scaleFactor ) {
        Matrix result = new MatrixSimple( iNoRows, iNoColumns );

        for (int i = 0; i < iNoRows; i++)
        for (int j = 0; j < iNoColumns; j++)
             result.setElement( i, j, scaleFactor.mult( this.getElement(i,j)) );

        return result;
    }

    // ====================================================================
    // Create and return the transpose of the invoking matrix
    // ====================================================================

    public Matrix transpose() {
        Matrix A = new MatrixSimple( iNoColumns, iNoRows );

        for (int i = 0; i < iNoRows; i++)
            for (int j = 0; j < iNoColumns; j++)
                A.setData(j,i, this.data[i][j] );
        return A;
    }

    // ====================================================================
    // return C = A + B
    // ====================================================================

    public Matrix add( Matrix B ) {

        // Check compatibility of matrix dimensions ....

        if (B.getNoRows() != this.iNoRows || B.getNoColumns() != this.iNoColumns ) throw
            new RuntimeException("*** In MatrixSimple.add(): Incompatible matrix dimensions ...");

        // Create new matrix 

        Matrix C = new MatrixSimple( iNoRows, iNoColumns );

        for (int i = 0; i < iNoRows; i++)
            for (int j = 0; j < iNoColumns; j++) {
                Quantity X = this.getElement(i,j);
                Quantity Y = B.getElement(i,j);
                C.setElement( i, j, X.add(Y) ); 
        }

        return C;
    }

    // ====================================================================
    // return C = A - B
    // ====================================================================

    public Matrix sub(Matrix B) {
        Matrix A = (Matrix) this;

        // Check compatibility of matrix dimensions ....

        if (B.getNoRows() != A.getNoRows() || B.getNoColumns() != A.getNoColumns() ) throw
            new RuntimeException("Illegal matrix dimensions.");

        Matrix C = new MatrixSimple( iNoRows, iNoColumns );

        for (int i = 0; i < iNoRows; i++)
            for (int j = 0; j < iNoColumns; j++) {
                Quantity X = this.getElement(i,j);
                Quantity Y = B.getElement(i,j);
                C.setElement( i, j, X.sub(Y) ); 
        }

        return C;
    }

    // ====================================================================
    // Test: does A = B exactly?
    // ====================================================================

    public boolean equals(MatrixSimple B) {
        MatrixSimple A = this;

        if (B.iNoRows != A.iNoRows || B.iNoColumns != A.iNoColumns ) throw
            new RuntimeException("Illegal matrix dimensions.");

        for (int i = 0; i < iNoRows; i++)
            for (int j = 0; j < iNoColumns; j++)
                if (A.data[i][j] != B.data[i][j]) return false;

        return true;
    }

    // ====================================================================
    // Compute C = A * B
    // ====================================================================

    public Matrix mult( Matrix B ) {

        if ( this.getNoColumns() != B.getNoRows() ) throw
             new RuntimeException("Illegal matrix dimensions.");

        Matrix C = new MatrixSimple( this.iNoRows, B.getNoColumns() );

        for (int i = 0; i < C.getNoRows(); i++) {
            for (int j = 0; j < C.getNoColumns(); j++) {
                for (int k = 0; k < this.getNoColumns(); k++) {
                    Quantity Y = this.getElement(i,k);
                    Quantity Z = B.getElement(k,j);
                    if ( k == 0 ) {
                       C.setElement( i, j, Y.mult(Z) ); 
                    } else {
                       Quantity X =   C.getElement(i,j);
                       C.setElement( i, j, X.add( Y.mult(Z) )); 
                    }
                }
            }
        }

        return C;
    }

    // ====================================================================
    // Compute C = A^n where n is a positive integer ...
    // ====================================================================

    public Matrix pow( double factor ) {
       Matrix result = null;
       int exponent = (int) factor;

       if ( exponent < 1 ) throw
          new RuntimeException("*** In MatrixSimple.pow(): Illegal matrix exponent.");
      
       result = this;
       for ( int i = 1; i < exponent; i = i + 1 )
           result = result.mult( this );

       return result;
    }

    // =========================================================
    // return x = A^-1 b, assuming A is square and has full rank
    // =========================================================

    public Matrix solve( Matrix rhs ) {

        if (this.iNoRows != this.iNoColumns ||
             rhs.getNoRows() != this.iNoColumns || rhs.getNoColumns() != 1)
                 throw new RuntimeException("Illegal matrix dimensions.");

        // Create copies of the data

        MatrixSimple A = (MatrixSimple) this.copy();
        MatrixSimple b = (MatrixSimple)  rhs.copy();

        // Gaussian elimination with partial pivoting

        for (int i = 0; i < iNoColumns; i++) {

            // find pivot row and swap

            int max = i;
            for (int j = i + 1; j < this.iNoColumns; j++)
                if (Math.abs(A.data[j][i]) > Math.abs(A.data[max][i]))
                    max = j;
            A.swap(i, max);
            b.swap(i, max);

            // singular

            if (A.data[i][i] == 0.0) throw
                new RuntimeException("MatrixSimple is singular.");

            // pivot within b

            for (int j = i + 1; j < iNoColumns; j++)
                b.data[j][0] -= b.data[i][0] * A.data[j][i] / A.data[i][i];

            // pivot within A

            for (int j = i + 1; j < iNoColumns; j++) {
                double m = A.data[j][i] / A.data[i][i];
                for (int k = i+1; k < iNoColumns; k++) {
                    A.data[j][k] -= A.data[i][k] * m;
                }
                A.data[j][i] = 0.0;
            }
        }

        // back substitution

        Matrix x = new MatrixSimple(iNoColumns, 1);

        for (int j = iNoColumns - 1; j >= 0; j--) {
            double t = 0.0;
            for (int k = j + 1; k < iNoColumns; k++)
                t += A.getData(j,k) * x.getData(k,0);
            x.setData( j, 0, ((b.getData(j,0) - t) / A.getData(j,j)) );
        }

        return x;
    }

    // =========================================================
    // Create copy of a matrix object ...
    // =========================================================

    public Matrix copy() {
       Matrix c = new MatrixSimple ( this.iNoRows, this.iNoColumns );
       for (int i = 0; i < this.iNoRows; i = i + 1 )
       for (int j = 0; j < this.iNoColumns; j = j + 1 )
           c.setData( i,j, this.data[i][j] );

       return c;
    }

    // ============================================
    // print matrix to standard output
    // ============================================

    // Define NoColumns to be a "constant" .... 

    public final static int NoColumns = 6;

    public void print() {
       int ii, ij, ib;   
       int iNoBlocks;
       int iFirstColumn, iLastColumn;

       // Compute no of blocks of rows to be printed .....

       if( this.iNoColumns % NoColumns == 0 )
          iNoBlocks = this.iNoColumns/NoColumns;
       else
          iNoBlocks = this.iNoColumns/NoColumns + 1;

       // Loop over the number of blocks ...

       for( ib = 1; ib <= iNoBlocks; ib++ ) { 
          iFirstColumn = (ib-1)*NoColumns + 1;
          iLastColumn  = Math.min ( ib*NoColumns, this.iNoColumns );

          // Print title of matrix at top of each block ....
       
          if( this.sName != null ) 
              System.out.println("\nMatrix: " + this.sName );
          else 
              System.out.println("\nMatrix: Untitled");

          // Label row and column nos */

          System.out.printf ("row/col          ");
          for( ii = iFirstColumn; ii <= iLastColumn; ii++ )
              System.out.printf("       %3d   ", ii);
          System.out.printf("\n");

          // Print column units ....

          System.out.printf("        units    ");
          for( ii = iFirstColumn; ii <= iLastColumn; ii++ ) {
               String uName = (String) d2s.get( this.colUnits[ii-1] );
               if(uName != null )
                  System.out.printf("%10s   ", uName );
               else
                  System.out.printf("             ");
          }
          System.out.printf("\n");

          // Print contents of matrix  ....

          for( ii = 1; ii <= iNoRows; ii++ ) {
              System.out.printf(" %3d ", ii);
              System.out.printf("         ");
              for( ij  = iFirstColumn; ij <= iLastColumn; ij++)
                  System.out.printf(" %12.5e", this.data[ ii-1 ][ ij-1 ] );
              System.out.printf("\n");
          }
       }
    }

    // ============================================
    // Convert array to string format ....
    // ============================================

    public String toString() {
       StringBuffer buf = new StringBuffer();
       int ii, ij, ib;   
       int iNoBlocks;
       int iFirstColumn, iLastColumn;

       // Compute no of blocks of rows to be printed .....

       if( this.iNoColumns % NoColumns == 0 )
          iNoBlocks = this.iNoColumns/NoColumns;
       else
          iNoBlocks = this.iNoColumns/NoColumns + 1;

       // Loop over the number of blocks ...

       for( ib = 1; ib <= iNoBlocks; ib++ ) { 
          iFirstColumn = (ib-1)*NoColumns + 1;
          iLastColumn  = Math.min ( ib*NoColumns, this.iNoColumns );

          // Print title of matrix at top of each block ....
       
          if( this.sName != null ) 
              buf.append( String.format("\nMatrix: %s\n", this.sName ) );
          else 
              buf.append( String.format("\nMatrix: Untitled\n" ) );

          // Label row and column nos */

          buf.append( String.format ("row/col          ") );
          for( ii = iFirstColumn; ii <= iLastColumn; ii++ )
              buf.append( String.format("       %3d   ", ii));
          buf.append( String.format("\n") );

          // Print column units ....

          buf.append( String.format("        units    ") );
          for( ii = iFirstColumn; ii <= iLastColumn; ii++ ) {
               String uName = (String) d2s.get( this.colUnits[ii-1] );
               if(uName != null )
                  buf.append( String.format("%10s   ", uName ));
               else
                  buf.append( String.format("             "));
          }
          buf.append( String.format("\n") );

          // Print contents of matrix  ....

          for( ii = 1; ii <= iNoRows; ii++ ) {
              buf.append( String.format(" %3d ", ii));
              buf.append( String.format("         "));
              for( ij  = iFirstColumn; ij <= iLastColumn; ij++)
                  buf.append( String.format(" %12.5e", this.data[ ii-1 ][ ij-1 ] ) );
              buf.append( String.format("\n") );
          }
       }

       return buf.toString();
    }

    // ================================================
    // Exercise methods in the MatrixSimple class ....
    // ================================================

    public static void main(String[] args) {
    }
}
